@@ -3,12 +3,17 @@
 __all__ = ["Grid"]
 
 import numpy as np
+from importlib.util import find_spec
 from math import floor, ceil
+from pathlib import Path
+from warnings import warn
 
 from .display import shorten
 from .logger import get_logger
 from .spatial import is_same_crs
 
+mask_cache = {}
+
 
 def float32_is_also_float64(val):
     """Return True if float32 and float64 values are the same."""
@@ -20,6 +25,59 @@ def float32_is_also_float64(val):
     return val64 == val6432
 
 
+def get_modflow_model(model, model_name=None, logger=None):
+    """Return model object from str or Path."""
+    import flopy
+
+    if hasattr(model, "modelgrid"):
+        return model
+    elif not isinstance(model, (str, Path)):
+        raise TypeError("expected str, Path or model instance")
+    pth = Path(model).resolve()
+    if not pth.exists():
+        raise ValueError(f"cannot read path '{pth}'")
+    elif (
+            (pth.is_dir() and (pth / "mfsim.nam").is_file()) or
+            pth.name == "mfsim.nam"):
+        # MODFLOW 6
+        if pth.is_dir():
+            sim_ws = str(pth)
+        else:
+            sim_ws = str(pth.parent)
+        if logger is not None:
+            logger.info("reading mf6 simulation from '%s'", sim_ws)
+        sim = flopy.mf6.MFSimulation.load(
+            sim_ws=sim_ws, strict=False, verbosity_level=0,
+            load_only=["dis", "tdis"])
+        model_names = list(sim.model_names)
+        if model_name is None:
+            model_name = model_names[0]
+            msg = "a model name should be specified, "
+            if len(model_names) > 1:
+                msg += "mfsim.nam has %d models (%s); selecting first"
+                args = (len(model_names), model_names)
+            else:
+                msg += "selecting %r from mfsim.nam"
+                args = (model_name,)
+            if logger is None:
+                warn(msg % args, UserWarning, stacklevel=2)
+            else:
+                logger.warning(msg, *args)
+        elif model_name not in model_names:
+            raise KeyError(
+                f"model name {model_name} not found in {model_names}")
+        model = sim.get_model(model_name)
+        setattr(model, "tdis", sim.tdis)  # this is a bit of a hack
+        return model
+    elif pth.is_file():  # assume 'classic' MOFLOW file
+        model = flopy.modflow.Modflow.load(
+            pth.name, model_ws=str(pth.parent), load_only=["dis", "bas6"],
+            check=False, verbose=False)
+        return model
+    raise ValueError(
+        f"cannot determine how to read MODFLOW model '{pth}'")
+
+
 class Grid:
     """Grid information class to discritize a spatial domain on a grid.
 
@@ -269,6 +327,180 @@ class Grid:
         return cls(resolution=resolution, shape=shape, top_left=top_left,
                    projection=projection, logger=logger)
 
+    @classmethod
+    def from_modflow(
+            cls, model, model_name=None, projection=None, logger=None):
+        """Create grid information from a MODFLOW model.
+
+        Parameters
+        ----------
+        model : str, Path, flopy.modflow.Modflow, or flopy.mf6.mfmodel.MFModel
+            MODFLOW model specified either as a FloPy object, or path to
+            a MODFLOW file.
+        model_name : str or None (default)
+            Needed if MODFLOW 6 simulation has more than one model.
+        projection : str, default None
+            Coordinate reference system described as a string either as (e.g.)
+            EPSG:2193 or a WKT string. Default None will attempt to obtain
+            this from the model, otherwise it will be "".
+        logger : logging.Logger, optional
+            Logger to show messages.
+
+        Raises
+        ------
+        ModuleNotFoundError
+            If flopy is not installed.
+        """
+        if find_spec("flopy") is None:
+            raise ModuleNotFoundError("from_modflow requires flopy")
+        if logger is None:
+            logger = get_logger(cls.__class__.__name__)
+        logger.info("creating grid info from a MODFLOW model: %s", type(model))
+        mask_cache_key = (repr(model), model_name)
+        model = get_modflow_model(model, model_name, logger)
+        modelgrid = model.modelgrid
+        delr = modelgrid.delr[0]
+        delc = modelgrid.delc[0]
+        if not (modelgrid.delr == delr).all():
+            raise ValueError("model delr is not constant")
+        elif not (modelgrid.delc == delc).all():
+            raise ValueError("model delc is not constant")
+        elif delr != delc:
+            raise ValueError("model delr and delc are different")
+        if modelgrid.angrot != 0:
+            logger.error("rotated model grids are not supported")
+        top_left = (
+            modelgrid.xoffset,
+            modelgrid.yoffset + modelgrid.delc.sum())
+        if projection is None:
+            if isinstance(modelgrid.epsg, int):
+                projection = f"EPSG:{modelgrid.epsg}"
+            elif modelgrid.proj4 is not None:
+                projection = modelgrid.proj4
+            else:
+                projection = ""
+        # also cache mask while we are here
+        if hasattr(model, "bas6"):
+            domain = model.bas6.ibound.array
+        else:
+            domain = model.dis.idomain.array
+        mask = (domain == 0).all(0)
+        mask_cache[mask_cache_key] = mask
+        return cls(
+            resolution=delr, shape=modelgrid.top.shape, top_left=top_left,
+            projection=projection, logger=logger)
+
+    def array_from_array(self, grid, array, resampling=None):
+        """Return array from a different grid and array.
+
+        Parameters
+        ----------
+        grid : Grid
+            Grid for array input.
+        array : array_like
+            Array data to regrid. If 3D, the first dimension is the band.
+        resampling : rasterio.enums.Resampling, optional
+            Choose one from rasterio.enums.Resampling; default (None)
+            automatically selects the best method based on the relative grid
+            resolutions and data type.
+
+        Returns
+        -------
+        np.ma.array
+
+        Raises
+        ------
+        ModuleNotFoundError
+            If rasterio is not installed.
+        """
+        try:
+            import rasterio
+            from rasterio.enums import Resampling
+            from rasterio.warp import reproject
+        except ModuleNotFoundError:
+            raise ModuleNotFoundError("array_from_array requires rasterio")
+        if not isinstance(grid, Grid):
+            raise TypeError(
+                f"expected grid to be a Grid; found {type(grid)!r}")
+        elif not (hasattr(array, "ndim") and hasattr(array, "shape")):
+            raise TypeError(
+                f"expected array to be array_like; found {type(array)!r}")
+        elif not (array.ndim in (2, 3) and array.shape[-2:] == grid.shape):
+            raise ValueError("array has different shape than grid")
+
+        rel_res_diff = abs(
+            (grid.resolution - self.resolution) / self.resolution) * 100
+        if rel_res_diff == 0.0:
+            self.logger.info(
+                "source and destination have the same resolution %s",
+                self.resolution)
+        else:
+            self.logger.info(
+                "source resolution %s vs destination resolution "
+                "%s, which is a relative difference of %s%%",
+                grid.resolution, self.resolution, rel_res_diff)
+        if resampling is None:
+            is_floating = np.issubdtype(array.dtype, np.floating)
+            if rel_res_diff <= 10.0:
+                resampling = Resampling.nearest
+            elif grid.resolution > self.resolution:
+                if is_floating:
+                    resampling = Resampling.bilinear
+                else:
+                    resampling = Resampling.nearest
+            elif grid.resolution < self.resolution:
+                if is_floating:
+                    resampling = Resampling.average
+                else:
+                    resampling = Resampling.mode
+            else:
+                raise ValueError()
+        self.logger.info("using %s resampling method", resampling)
+        if not rasterio.dtypes.check_dtype(array.dtype):
+            dtype = rasterio.dtypes.get_minimum_dtype(array)
+            self.logger.debug(
+                "changing array dtype from '%s' to '%s'",
+                array.dtype, dtype)
+            array = array.astype(dtype)
+        kwds = {}
+        nodata = None
+        if np.ma.isMA(array):
+            if array.dtype == np.float32:
+                fill_value = array.fill_value
+                if not float32_is_also_float64(fill_value):
+                    # TODO: any better way to find fill_value?
+                    fill_value = 3.28e9
+                    assert float32_is_also_float64(fill_value)
+                    assert fill_value not in array
+                    array = array.copy()
+                    array.fill_value = fill_value
+            kwds["src_nodata"] = nodata = array.fill_value
+            array = array.filled()
+        src_crs = self.projection
+        dst_crs = grid.projection
+        if not src_crs and not dst_crs:
+            # TODO: is there a better catch-all projection?
+            src_crs = "EPSG:3857"
+            dst_crs = "EPSG:3857"
+        elif not src_crs:
+            src_crs = dst_crs
+        elif not dst_crs:
+            dst_crs = src_crs
+        if array.ndim == 3:
+            dst_shape = array.shape[0:1] + self.shape
+        else:
+            dst_shape = self.shape
+        dst_array = np.ma.empty(dst_shape, array.dtype)
+        dst_array.mask = False
+        _ = reproject(
+            array, dst_array.data,
+            src_transform=grid.transform, dst_transform=self.transform,
+            src_crs=src_crs, dst_crs=dst_crs,
+            resampling=resampling, **kwds)
+        if nodata is not None:
+            dst_array.mask = dst_array.data == nodata
+        return dst_array
+
     def array_from_raster(self, fname: str, bidx: int = 1, resampling=None):
         """Return array from a raster source aligned to grid info.
 
@@ -312,7 +544,6 @@ class Grid:
 
             band = rasterio.band(ds, bidx)
             ar = np.ma.zeros(self.shape, band.dtype)
-            is_floating = np.issubdtype(ar.dtype, np.floating)
             ds_mean_res = np.mean(ds.res)
             rel_res_diff = abs(
                 (ds_mean_res - self.resolution) / self.resolution) * 100
@@ -326,6 +557,7 @@ class Grid:
                     "%s, which is a relative difference of %s%%",
                     ds_mean_res, self.resolution, rel_res_diff)
             if resampling is None:
+                is_floating = np.issubdtype(ar.dtype, np.floating)
                 if rel_res_diff <= 10.0:
                     resampling = Resampling.nearest
                 elif ds_mean_res > self.resolution:
@@ -415,6 +647,7 @@ class Grid:
             from rasterio.dtypes import get_minimum_dtype
             from rasterio.enums import Resampling
             from rasterio.warp import reproject
+            from shapely.geometry import shape
         except ModuleNotFoundError:
             raise ModuleNotFoundError(
                 "array_from_vector requires fiona and rasterio")
@@ -447,7 +680,7 @@ class Grid:
             else:
                 do_transform = True
                 from fiona.transform import transform_geom
-                from shapely.geometry import box, mapping, shape
+                from shapely.geometry import box, mapping
                 self.logger.info(
                     "geometries will be transformed from %s to %s",
                     shorten(ds_crs, 60), shorten(grid_crs, 60))
@@ -464,6 +697,7 @@ class Grid:
                 if attribute not in attributes:
                     raise KeyError(
                         f"could not find {attribute} in {attributes}")
+                vdtype = ds.schema["properties"][attribute]
             if do_transform:
                 grid_box = box(*grid_bounds)
                 # TODO: does this make sense?
@@ -475,38 +709,58 @@ class Grid:
                 self.logger.info(
                     "transforming features in bbox %s", grid_box_t.bounds)
                 for _, feat in ds.items(**kwargs):
-                    geom = transform_geom(ds_crs, grid_crs, feat["geometry"])
-                    if not grid_box.intersects(shape(geom)):
-                        continue
                     if attribute is not None:
                         val = feat["properties"][attribute]
+                        if val is None:
+                            continue
+                    geom = transform_geom(ds_crs, grid_crs, feat["geometry"])
+                    geom_obj = shape(geom)
+                    if geom_obj.is_empty or not grid_box.intersects(geom_obj):
+                        continue
+                    if attribute is not None:
                         vals.append(val)
                     geom_vals.append((geom, val))
             else:
                 for _, feat in ds.items(bbox=grid_bounds):
-                    geom = feat["geometry"]
                     if attribute is not None:
                         val = feat["properties"][attribute]
+                        if val is None:
+                            continue
+                    geom = feat["geometry"]
+                    if shape(geom).is_empty:
+                        continue
+                    if attribute is not None:
                         vals.append(val)
                     geom_vals.append((geom, val))
-            if attribute is not None:
-                vals = np.array(vals)
-                vdtype = ds.schema["properties"][attribute]
-                if vdtype.startswith("float"):
-                    nodata = vals.max() * 10.0
-                    resampling = Resampling.average
-                elif vdtype.startswith("int"):
-                    if vals.min() > 0:
-                        nodata = 0
-                    else:
-                        nodata = vals.max() + 1
-                    resampling = Resampling.mode
+
+        if len(geom_vals) == 0:
+            self.logger.warning("no valid geometry objects found")
+            nodata = 0
+            if attribute is not None and vdtype.startswith("float"):
+                nodata = 0.0
+            dtype = get_minimum_dtype(nodata)
+            ar = np.ma.zeros(self.shape, dtype=dtype)
+            ar.mask = True
+            ar.fill_value = fill
+            return ar
+
+        if attribute is not None:
+            vals = np.array(vals)
+            if vdtype.startswith("float"):
+                nodata = vals.max() * 10.0
+                resampling = Resampling.average
+            elif vdtype.startswith("int"):
+                if vals.min() > 0:
+                    nodata = 0
                 else:
-                    raise ValueError(
-                        f"attribute {attribute} is neither float or int")
-                dtype = get_minimum_dtype(np.append(vals, nodata))
-                if dtype == "float32":
-                    dtype = "float64"
+                    nodata = vals.max() + 1
+                resampling = Resampling.mode
+            else:
+                raise ValueError(
+                    f"attribute {attribute} is neither float or int")
+            dtype = get_minimum_dtype(np.append(vals, nodata))
+            if dtype == "float32":
+                dtype = "float64"
         dtype_conv = np.dtype(dtype).type
         nodata = dtype_conv(nodata)
         fill = dtype_conv(fill)
@@ -557,7 +811,35 @@ class Grid:
         ar = self.array_from_vector(fname, None, layer=layer)
         return ~ar.data.astype(bool)
 
-    def write_raster(self, array, fname: str, driver: str = "GTiff"):
+    def mask_from_modflow(self, model, model_name=None):
+        """Return a mask array from a MODFLOW model, based on IBOUND/IDOMAIN.
+
+        Parameters
+        ----------
+        model : str, Path, flopy.modflow.Modflow, or flopy.mf6.mfmodel.MFModel
+            MODFLOW model specified either as a FloPy object, or path to
+            a MODFLOW file.
+        model_name : str or None (default)
+            Needed if MODFLOW 6 simulation has more than one model.
+
+        Returns
+        -------
+        np.array
+        """
+        mask_cache_key = (repr(model), model_name)
+        if mask_cache_key in mask_cache:
+            self.logger.info("using mask from cache")
+            return mask_cache[mask_cache_key]
+        model = get_modflow_model(model, model_name, self.logger)
+        if hasattr(model, "bas6"):
+            domain = model.bas6.ibound.array
+        else:
+            domain = model.dis.idomain.array
+        mask = (domain == 0).all(0)
+        mask_cache[mask_cache_key] = mask
+        return mask
+
+    def write_raster(self, array, fname: str, driver=None):
         """Write array to a raster file format.
 
         Parameters
@@ -566,8 +848,8 @@ class Grid:
             Array to write; must have 2-dimensions that match shape.
         fname : str
             Output file to write.
-        driver : str, optional
-            Raster driver. Default is "GTiff" for GeoTIFF.
+        driver : str or None (default)
+            Raster driver. Default None will determine driver from fname.
 
         Raises
         ------
@@ -583,6 +865,9 @@ class Grid:
         elif array.shape != self.shape:
             raise ValueError("array must have same shape " + str(self.shape))
         self.logger.info("writing raster file: %s", fname)
+        if driver is None:
+            from rasterio.drivers import driver_from_extension
+            driver = driver_from_extension(fname)
         kwds = {
             "driver": driver,
             "width": self.shape[1],
